#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2014 by Serge Poltavski                                 #
# serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #


__author__ = 'Serge Poltavski'

from pdobject import *
from pdcomment import PdComment
import common


class PdCanvas(PdObject):
    TYPE_NONE, TYPE_WINDOW, TYPE_SUBPATCH, TYPE_GRAPH = range(0, 4)

    def __init__(self, x, y, w, h, **kwargs):
        PdObject.__init__(self, "cnv", x, y, w, h)
        self._objects = []
        self._id_counter = 0
        self._name = ""
        self._type = self.TYPE_NONE
        self._graphs = []
        self._connections = {}

        if 'name' in kwargs:
            self._name = kwargs['name']

        if 'open_on_load' in kwargs:
            self.open_on_load = kwargs['open_on_load']

    @property
    def objects(self):
        return self._objects

    @property
    def graphs(self):
        return self._graphs

    @property
    def connections(self):
        return self._connections

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, t):
        assert t in (self.TYPE_NONE, self.TYPE_WINDOW, self.TYPE_SUBPATCH, self.TYPE_GRAPH)
        self._type = t

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    def append_graph(self, obj):
        assert self != obj
        assert isinstance(obj, PdCanvas)
        assert obj.type == self.TYPE_GRAPH
        self.append_object(obj)

    def gen_object_id(self):
        res = self._id_counter
        self._id_counter += 1
        return res

    def append_object(self, obj):
        assert self != obj
        assert issubclass(obj.__class__, PdBaseObject)

        if obj in self._objects:
            common.warning(u"object already on canvas: {0:s}".format(obj))
            return False

        obj.id = self.gen_object_id()
        self.objects.append(obj)
        return True

    def find_object_by_id(self, oid):
        for obj in self._objects:
            if issubclass(obj.__class__, PdObject):
                if obj.id == oid:
                    return obj

        return None

    @staticmethod
    def make_connection_key(sid, soutl, did, dinl):
        return "%i:%i => %i:%i" % (sid, soutl, did, dinl)

    def add_connection(self, sid, soutl, did, dinl, check_xlets=True):
        assert sid != did
        assert sid >= 0 and did >= 0 and soutl >= 0 and dinl >= 0

        def str_bbox(obj, xlet):
                if obj:
                    return "[%s:%d]" % (obj.name, xlet)
                else:
                    return "[?:%d]" % xlet

        src_obj = self.find_object_by_id(sid)
        dest_obj = self.find_object_by_id(did)

        if src_obj and dest_obj:

            if check_xlets:
                if soutl >= len(src_obj.outlets()):
                    common.warning(u"[{0:s}] invalid outlet number: {1:d}, total outlets: {2:d}".
                                   format(src_obj.name, soutl, len(src_obj.outlets())))
                    return False

                if dinl >= len(dest_obj.inlets()):
                    common.warning(u"[{0:s}] invalid inlet number: {1:d}, total inlets: {2:d}".
                                   format(dest_obj.name, dinl, len(dest_obj.inlets())))
                    return False

            ckey = PdCanvas.make_connection_key(sid, soutl, did, dinl)
            self._connections[ckey] = (src_obj, soutl, dest_obj, dinl)
            return True
        else:
            common.warning("%s => %s" % (str_bbox(src_obj, soutl), str_bbox(dest_obj, dinl)))
            # common.warning("[%s:%d] => [%s:%d]" % (src_obj.name, soutl, dest_obj.name, dinl))
            common.warning("connection not found: %s:%d => %s:%d in canvas: %s" % (sid, soutl, did, dinl, self.name))
            return False

    def connect(self, args):
        assert len(args) == 4
        src_id = int(args[0])
        src_outl = int(args[1])
        dest_id = int(args[2])
        dest_inl = int(args[3])

        self.add_connection(src_id, src_outl, dest_id, dest_inl)

    def append_subpatch(self, obj):
        assert self != obj
        assert isinstance(obj, PdCanvas)
        assert obj.type == self.TYPE_SUBPATCH
        self.append_object(obj)
        return True

    def __str__(self):
        if self.type == self.TYPE_WINDOW:
            name = "Canvas "
        elif self.type == self.TYPE_GRAPH:
            name = "Graph "
        elif self.type == self.TYPE_SUBPATCH:
            name = "Subpatch "
        else:
            print self.name
            assert False

        name += u"\"{0:s}\"".format(self.name)

        res = "%-30s (%i,%i %ix%i id:%i)\n" % (name, self._x, self._y, self._width, self._height, self.id)
        for obj in self.objects:
            res += "    " + str(obj)
            res += "\n"

        return res

    def draw(self, painter):
        if self.type == self.TYPE_NONE:
            return

        if self.type == self.TYPE_SUBPATCH:
            painter.draw_subpatch(self)
        elif self.type == self.TYPE_GRAPH:
            painter.draw_graph(self)
        elif self.type == self.TYPE_WINDOW:
            painter.draw_canvas(self)

            for obj in self._objects:
                obj.draw(painter)

            painter.draw_connections(self)

    def inlets(self):
        res = []

        objects = sorted(self._objects, key=lambda obj: obj.x)

        for o in objects:
            if issubclass(o.__class__, PdObject):
                if o.name == "inlet":
                    res.append(self.XLET_MESSAGE)
                elif o.name == "inlet~":
                    res.append(self.XLET_SOUND)

        return res

    def outlets(self):
        res = []

        objects = sorted(self._objects, key=lambda obj: obj.x)

        for o in objects:
            if issubclass(o.__class__, PdObject):
                if o.name == "outlet":
                    res.append(self.XLET_MESSAGE)
                elif o.name == "outlet~":
                    res.append(self.XLET_SOUND)

        return res

    def traverse(self, visitor):
        if hasattr(visitor, 'visit_canvas_begin'):
            visitor.visit_canvas_begin(self)

        for o in self.objects:
            o.traverse(visitor)

        for k, conn in self._connections.items():
            visitor.visit_connection(conn)

        if hasattr(visitor, 'visit_canvas_end'):
            visitor.visit_canvas_end(self)

    def _print_connections(self):
        for k, v in self.connections.items():
            print "[%s:%d (%d,%d)] => [%s:%d (%d,%d)]" % (v[0].name, v[1], v[0].x, v[0].y,
                                                          v[2].name, v[3], v[2].x, v[2].y)


#!/usr/bin/env python
# coding=utf-8
# Copyright (C) 2014 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

from pddrawstyle import *
import cairo
import textwrap
from pdcomment import *
from pdobject import *
from pdmessage import *


class BRectCalculator(object):
    def __init__(self):
        self._style = PdDrawStyle()
        self._ims = cairo.ImageSurface(cairo.FORMAT_ARGB32, 10, 10)
        self._cr = cairo.Context(self._ims)

        self.st_font_slant = cairo.FONT_SLANT_NORMAL
        self.st_font_weight = cairo.FONT_WEIGHT_NORMAL
        self.st_line_join = cairo.LINE_JOIN_ROUND
        self._cr.select_font_face(self._style.font_family, self.st_font_slant, self.st_font_weight)
        self._cr.set_font_size(self._style.font_size)

        self._bboxes = []

    def clear(self):
        self._bboxes = []

    @property
    def bboxes(self):
        return self._bboxes

    def brect(self):
        left = min(self._bboxes, key=lambda o: o[0])[0]
        top = min(self._bboxes, key=lambda o: o[1])[1]
        right = max(self._bboxes, key=lambda o: o[0] + o[2])
        right = right[0] + right[2]
        bottom = max(self._bboxes, key=lambda o: o[1] + o[3])
        bottom = bottom[1] + bottom[3]

        return left, top, right - left, bottom - top

    def object_brect(self, obj):
        assert isinstance(obj, PdObject)
        txt = obj.to_string()
        (x, y, width, height, dx, dy) = self._cr.text_extents(txt)

        w = max(width + self._style.obj_pad_x * 2, self._style.obj_min_width)
        h = self._style.obj_height
        return obj.x, obj.y, w, h

    def text_brect(self, text):
        lines = []
        for line in textwrap.wrap(text, 59):
            line = ";\n".join(line.split(";")).split("\n")
            for subl in line:
                if subl:
                    lines.append(subl.strip())

        maxwd = 0
        for line in lines:
            maxwd = max(maxwd, self._cr.text_extents(line)[2])

        height = len(lines) * self._style.font_size
        return 0, 0, maxwd, height

    def comment_brect(self, comment):
        assert isinstance(comment, PdComment)
        return self.text_brect(comment.text())

    def message_brect(self, message):
        assert  isinstance(message, PdMessage)
        (x, y, width, height, dx, dy) = self._cr.text_extents(message.to_string())
        w = width + self._style.obj_pad_x * 4
        h = self._style.obj_height
        return message.x, message.y, w, h

    def visit_object(self, obj):
        bbox = self.object_brect(obj)
        obj.set_width(bbox[2])
        obj.set_height(bbox[3])
        self._bboxes.append(bbox)

    def visit_message(self, msg):
        bbox = self.message_brect(msg)
        msg.set_width(bbox[2])
        msg.set_height(bbox[3])
        self._bboxes.append(bbox)

    def visit_comment(self, comment):
        (x, y, w, h) = self.comment_brect(comment)
        self._bboxes.append((comment.x, comment.y, w, h))

    def visit_core_gui(self, gui):
        pass

    def visit_graph(self, g):
        pass

    def visit_connection(self, c):
        pass


#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

from pdfloatatom import PdFloatAtom
from pdcoregui import PdCoreGui
from pdobject import PdObject
from pdbng import PdBng
from pdtoggle import PdToggle
from pdslider import PdSlider
from pdradio import PdRadio
from pdgcanvas import PdGCanvas
import os
import re
import imp
import sys

externals = {}


def make(atoms):
    assert isinstance(atoms, list)
    assert len(atoms) > 0
    name = atoms[0]

    if name == "floatatom":
        return PdFloatAtom.from_atoms(atoms[1:])
    elif name == "bng":
        return PdBng.from_atoms(atoms[1:])
    elif name == "tgl":
        return PdToggle.from_atoms(atoms[1:])
    elif name in ("hsl", "vsl"):
        return PdSlider.from_atoms(atoms)
    elif name in ("hradio", "vradio"):
        return PdRadio.from_atoms(atoms)
    elif name == "cnv":
        return PdGCanvas.from_atoms(atoms[1:])
    # elif name in ("nbx", "vu"):
    #     return PdCoreGui(name, 0, 0, atoms[1:])
    elif find_external_object(name):
        return externals[name].create(atoms)
    else:
        return PdObject(name, 0, 0, 0, 0, atoms[1:])


def make_by_name(name, args=[], **kwargs):
    if name == "floatatom":
        return PdFloatAtom(0, 0, **kwargs)
    elif name == "bng":
        return PdBng(0, 0, **kwargs)
    else:
        return PdObject(name, 0, 0, 0, 0, args)


def find_external_object(name):
    rname = re.compile(r"^([a-z0-9~/]+)$")
    if not rname.match(name):
        return False

    mod_path = os.path.dirname(__file__) + "/externals/" + name
    if not os.path.exists(mod_path + ".py"):
        return False

    mod_dir = os.path.dirname(mod_path)
    mod_name = os.path.basename(mod_path)

    if mod_dir not in sys.path:
        sys.path.append(mod_dir)
    try:
        mod = __import__(mod_name)
        externals[name] = mod
        return True
    except ImportError:
        return None



#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2014 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                            #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

import os
import pdcanvas
import pddrawer
import pdobject
import cairopainter
import pdparser
from layout import *
from brectcalculator import *
from pdcomment import *
from pdmessage import *
from pdobject import *
from mako.template import Template
from mako.lookup import TemplateLookup
from docobject import *
from pdlayout import *
import pdfactory


class HtmlDocVisitor(object):
    def __init__(self):
        self._title = ""
        self._description = ""
        self._keywords = []
        self._website = ""
        self._library = ""
        self._category = ""
        self._version = ""
        self._license = {}
        self._aliases = []
        self._see_also = []
        self._examples = []
        self._authors = []
        self._contacts = ""
        self._inlets = {}
        self._outlets = {}
        self._arguments = []
        self._inlet_counter = 0
        self._image_counter = 0
        self._css_theme = "../theme.css"
        self._img_output_dir = "./out"
        # template config
        tmpl_path = "{0:s}/html_object.tmpl".format(os.path.dirname(__file__))
        # self._tmpl_lookup = TemplateLookup(directories=[os.path.dirname(__file__)])
        self._html_template = Template(filename=tmpl_path)
        self._layout = PdLayout()
        self._canvas_padding = 10

    def title_begin(self, t):
        self._title = t.text()

    def website_begin(self, w):
        self._website = w.text()

    def keywords_begin(self, k):
        self._keywords = k.keywords()

    def description_begin(self, d):
        self._description = d.text()

    def aliases_begin(self, a):
        self._aliases += a.aliases()

    def license_begin(self, l):
        self._license['url'] = l.url()
        self._license['name'] = l.name()

    def library_begin(self, lib):
        self._library = lib.text()

    def category_begin(self, cat):
        self._category = cat.text()

    def version_begin(self, v):
        self._version = v.text()

    def author_begin(self, author):
        self._authors.append(author.text())

    def contacts_begin(self, cnt):
        self._contacts = cnt.text()

    def see_begin(self, see):
        dict = {'name' : see.text()}
        dict['image'] = "object_{0:s}.png".format(see.text())
        dict['link'] = "{0:s}.html".format(see.text())
        self._see_also.append(dict)

    def make_image_id_name(self):
        self._image_counter += 1
        return self._image_counter, "image_{0:02d}.png".format(self._image_counter)

    def pdexample_begin(self, pd):
        if pd.file():
            parser = pdparser.PdParser()
            parser.parse(pd.file())
            self._layout.canvas = parser.canvas
        else:
            self._layout.canvas = pdcanvas.PdCanvas(0, 0, 10, 10, name="10")
            self._layout.canvas.type = pdcanvas.PdCanvas.TYPE_WINDOW

    def pdexample_end(self, pd):
        w, h = self.draw_area_size(pd)
        img_id, img_fname = self.make_image_id_name()

        example_dict = {'id' : img_id, 'image' : img_fname, 'title' : pd.title(), 'file' : pd.file() }

        if not pd.file():
            self._layout.update()

        self._examples.append(example_dict)

        output_fname = "./out/" + img_fname
        painter = cairopainter.CairoPainter(w, h, output_fname,
                                            xoffset=self._canvas_padding,
                                            yoffset=self._canvas_padding)
        walker = pddrawer.PdDrawer()
        walker.draw(self._layout.canvas, painter)

    def pdcomment_begin(self, comment):
        self._layout.comment(comment)

    def row_begin(self, row):
        self._layout.row_begin()

    def row_end(self, row):
        self._layout.row_end()

    def col_begin(self, col):
        self._layout.col_begin()

    def col_end(self, col):
        self._layout.col_end()

    def pdmessage_begin(self, msg_obj):
        self._layout.message_begin(msg_obj)

    def pdobject_begin(self, doc_obj):
        self._layout.object_begin(doc_obj)

    def draw_area_size(self, pd):
        assert isinstance(pd, DocPdexample)
        w = 0
        h = 0

        if pd.file() and pd.size() == "canvas":
            w, h = self._layout.canvas_brect()[2:]
        elif pd.size() == "auto":
            w, h = self._layout.layout_brect()[2:]
        elif not pd.size():
            w = pd.width()
            h = pd.height()

            if not w:
                w = self._layout.layout_brect()[2]
            if not h:
                h = self._layout.layout_brect()[3]
        else:
            w, h = self._layout.layout_brect()[2:]

        return int(w + 2 * self._canvas_padding), int(h + 2 * self._canvas_padding)

    def pdconnect_begin(self, c):
        self._layout.connect_begin(c)

    def inlets_begin(self, inlets):
        self._inlets = inlets.inlet_dict()

    def outlets_begin(self, outlets):
        self._outlets = outlets.outlet_dict()

    def arguments_begin(self, args):
        self._arguments = args.items()

    def generate_object_image(self, name):
        fname = "out/object_{0:s}.png".format(name)
        if os.path.exists(fname):
            return

        pdo = pdfactory.make_by_name(name)
        brect = BRectCalculator().object_brect(pdo)
        pad = 1  # pixel
        painter = cairopainter.CairoPainter(int(brect[2]) + pad, int(brect[3]) + pad, fname, "png")
        pdo.draw(painter)

    def generate_images(self):
        if self._aliases:
            for a in [self._title] + self._aliases:
                self.generate_object_image(a)

        if self._see_also:
            for sa in self._see_also:
                self.generate_object_image(sa['name'])

    def render(self):
        return self._html_template.render(
            title=self._title,
            description=self._description,
            keywords=self._keywords,
            image_dir='.',
            css_theme=self._css_theme,
            aliases=[self._title] + self._aliases,
            license=self._license,
            version=self._version,
            examples=self._examples,
            inlets=self._inlets,
            outlets=self._outlets,
            arguments=self._arguments,
            see_also=self._see_also,
            website=self._website,
            authors=self._authors,
            contacts=self._contacts,
            library=self._library,
            category=self._category)


#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

 
__author__ = 'Serge Poltavski'


from pdobject import PdObject


class PdArray(PdObject):
    def __init__(self, name, size, save=0):
        PdObject.__init__(self, name)
        self._size = int(size)
        self._save = save
        self._data = []
        self._xrange = (0, self._size)
        self._yrange = (-1, +1)

    def set(self, atoms):
        assert len(atoms) == self._size
        self._data = []
        for a in atoms:
            self._data.append(float(a))

    def set_xrange(self, min, max):
        self._xrange = (min, max)

    def set_yrange(self, min, max):
        self._yrange = (min, max)

    @staticmethod
    def from_atoms(atoms):
        assert len(atoms) == 4
        return PdArray(atoms[0], atoms[1], atoms[3])

    def draw(self, painter):
        painter.draw_rect(*self.brect())
        painter.draw_text(self.x, self.y - 5, self.name)

        if not self._data:
            return
        
        swd = self.width / self._size
        yrange = self._yrange[1] - self._yrange[0]
        sht = self.height / yrange
        for idx in xrange(0, self._size):
            x0 = self.left + swd * idx
            x1 = x0 + swd
            y = self.top + sht * self._data[idx] + (self.height * self._yrange[1])/yrange
            painter.draw_line(x0, y, x1, y, width=3, color=(0, 0, 0))


# coding=utf-8
from __future__ import print_function

#   Copyright (C) 2014 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #


__author__ = 'Serge Poltavski'

import sys
import os.path as path
from termcolor import colored
from colorama import init
import inspect

# use Colorama to make Termcolor work on Windows too
init()


def warning(*objs):
    print(colored("WARNING: ", "red"), *objs, file=sys.stderr)


def info(*objs):
    print(colored("INFO: ", "yellow"), *objs, file=sys.stdout)


def error_place():
    info = inspect.getframeinfo(inspect.currentframe().f_back)[0:3]
    return '[file: %s, method: %s, line:%d]' % (path.basename(info[0]), info[2], info[1])


#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

from pdcoregui import PdCoreGui
from brectcalculator import BRectCalculator


class PdFloatAtom(PdCoreGui):
    _bcalc = BRectCalculator()

    def __init__(self, x, y, **kwargs):
        PdCoreGui.__init__(self, "floatatom", x, y, [])
        self._digits = int(kwargs.get("digits", 5))
        assert self._digits >= 0

        self._min = kwargs.get("min", 0)
        self._max = kwargs.get("max", 0)
        assert self._min <= self._max

        self._label_pos = self.POS_LEFT
        if "label_pos" in kwargs:
            pos = self._get_label_pos(kwargs["label_pos"])
            if pos is not None:
                self._label_pos = pos

        self.label = kwargs.get("label", "-")
        self.send = kwargs.get("send", "-")
        self.receive = kwargs.get("receive", "-")

        self._value = float(kwargs.get("value", 0))

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v):
        assert isinstance(v, float) or isinstance(v, int)
        self._value = v

    @staticmethod
    def from_atoms(atoms):
        assert isinstance(atoms, list)
        assert len(atoms) == 9

        x = atoms[0]
        y = atoms[1]

        return PdFloatAtom(x, y,
                           digits=int(atoms[2]),
                           min=int(atoms[3]),
                           max=int(atoms[4]),
                           label_pos=int(atoms[5]),
                           label=atoms[6],
                           send=atoms[7],
                           receive=atoms[8])

    def get_height(self):
        return 17

    def get_width(self):
        return max(self._digits * 7, 20)

    def no_send(self):
        return self.send == "-"

    def no_receive(self):
        return self.receive == "-"

    def has_label(self):
        return self._label and self._label != "-"

    def outlets(self):
        if self.no_send():
            return [self.XLET_MESSAGE]
        else:
            return []

    def inlets(self):
        if self.no_receive():
            return [self.XLET_MESSAGE]
        else:
            return []

    def _is_int(self):
        return abs(self._value - int(self._value)) < 0.001

    def _str_value(self):
        if self._is_int():
            return str(int(self._value))
        else:
            return "{0:.2f}".format(self.value)

    def _label_brect(self):
        if not self.has_label():
            return 0, 0, 0, 0
        return self._bcalc.text_brect(self._label)

    def draw(self, painter):
        vertexes = (
            (self.x, self.y),
            (0, self.height),
            (self.width, 0),
            (0, -10),
            (-7, -7)
        )

        text_yoff = 13

        painter.draw_poly(vertexes, fill=(0.88, 0.88, 0.88), outline=(0.75, 0.75, 0.75))
        painter.draw_text(self.x + 2, self.top + text_yoff, self._str_value())
        painter.draw_inlets(self.inlets(), self.x, self.y, self.width)
        painter.draw_outlets(self.outlets(), self.x, self.bottom, self.width)

        if self.has_label():
            lx, ly, lw, lh = self._label_brect()
            pad = 3
            if self._label_pos == self.POS_LEFT:
                lx = self.x - lw - pad
                ly = self.y + text_yoff
            elif self._label_pos == self.POS_RIGHT:
                lx = self.right + pad
                ly = self.y + text_yoff
            elif self._label_pos == self.POS_BOTTOM:
                lx = self.left
                ly = self.bottom + text_yoff
            elif self._label_pos == self.POS_TOP:
                lx = self.left
                ly = self.top - pad
            else:
                assert False

            painter.draw_text(lx, ly, self._label)


#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2014 by Serge Poltavski                                  #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'


class PdPainter(object):
    def draw_canvas(self, canvas):
        pass

    def draw_comment(self, comment):
        print "Draw comment: #", comment.text()

    def draw_message(self, message):
        print "Draw message [id:%i]: %s" % (message.id, message.to_string())

    def draw_object(self, obj):
        print "Draw object: [id:%i] [%s]" % (obj.id, " ".join(obj.args))

    def draw_core_gui(self, gui):
        print "Draw core GUI: [id:%i] [%s]" % (gui.id, gui.name)

    def draw_subpatch(self, subpatch):
        print u"Draw subpatch: [pd {0:s}]".format(subpatch.name)

    def draw_graph(self, graph):
        print "Draw graph "

    def draw_connections(self, canvas):
        print "Draw connections "

    def draw_poly(self, vertexes, **kwargs):
        print "Draw poly:", vertexes

    def draw_text(self, x, y, text, **kwargs):
        print "Draw text:", text

    def draw_inlets(self, inlets, x, y, width):
        print "Draw inlets:", inlets

    def draw_outlets(self, outlets, x, y, width):
        print "Draw outlets:", outlets

    def draw_circle(self, x, y, width, **kwargs):
        print "Draw circle"

    def draw_line(self, x0, y0, x1, y1, **kwargs):
        print "Draw line"

    def draw_rect(self, x, y, w, h, **kwargs):
        print "Draw rect"


#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2014 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #


__author__ = 'Serge Poltavski'

from pdbaseobject import *


class PdComment(PdBaseObject):
    def __init__(self, x, y, args):
        super(PdComment, self).__init__(x, y, 0, 0)
        self.args = args

    @staticmethod
    def unescape(s):
        return s.strip() \
            .replace(chr(13), "") \
            .replace(chr(10), "") \
            .replace("\\;", ";") \
            .replace("\\,", ",")

    def text(self):
        res = ""

        for a in self.args:
            a = self.unescape(a)

            if a == ",":
                res += ","
            elif a == ";":
                res += ";"
            else:
                res += " " + a

        res = res.strip()

        return res

    def __str__(self):
        res = "# %-39s {x:%i,y:%i}" % (self.text(), self._x, self._y)
        return res

    def draw(self, painter):
        painter.draw_comment(self)

    def traverse(self, visitor):
        visitor.visit_comment(self)


#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2014 by Serge Poltavski                                   #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

import six

from pdbaseobject import *
from xletcalcdatabase import *


class PdObject(PdBaseObject):
    xlet_calculator = XletCalcDatabase()
    # calculates xlets number
    XMETHOD_CALCULATE = 0
    # no calculation, xlets should be defined manually by calling set_inlets
    XMETHOD_EXPLICIT = 1

    def __init__(self, name, x=0, y=0, w=0, h=0, args=[]):
        PdBaseObject.__init__(self, x, y, w, h)
        assert isinstance(name, six.string_types)
        assert len(name) > 0

        self._name = name

        assert isinstance(args, list)
        self._args = args

        self._inlets = []
        self._outlets = []
        self._xlets_method = self.XMETHOD_CALCULATE

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, n):
        assert isinstance(n, six.string_types)
        assert len(n) > 0
        self._name = n

    @property
    def args(self):
        return self._args

    def append_arg(self, a):
        self._args.append(a)

    def num_args(self):
        return len(self._args)

    def args_to_string(self):
        res = ""

        esc_args = []
        for arg in self.args:
            esc_args.append(PdObject.unescape(arg))

        for arg in esc_args:
            if arg == ",":
                res += ", "
            elif arg == ";":
                res += ";"
            else:
                res += " " + arg

        res = " ".join(res.strip().split())
        return res

    def to_string(self):
        res = PdObject.unescape(self.name) + ' ' + self.args_to_string()
        return res.strip()

    def __str__(self):
        res = "[%-40s {x:%i,y:%i,id:%i}" % (self.to_string() + "]", self.x, self.y, self.id)
        return res

    def draw(self, painter):
        painter.draw_object(self)

    def inlets(self):
        if self._xlets_method == self.XMETHOD_EXPLICIT:
            return self._inlets
        elif self._xlets_method == self.XMETHOD_CALCULATE:
            return PdObject.xlet_calculator.inlets(self)
        else:
            return []

    def outlets(self):
        if self._xlets_method == self.XMETHOD_EXPLICIT:
            return self._outlets
        elif self._xlets_method == self.XMETHOD_CALCULATE:
            return PdObject.xlet_calculator.outlets(self)
        else:
            return []

    def traverse(self, visitor):
        visitor.visit_object(self)


#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

 
__author__ = 'Serge Poltavski'

from pdcoregui import PdCoreGui


class PdGCanvas(PdCoreGui):
    # X [size]? [width]? [height]? [send]? [receive]? [label]?
    # [x_off]? [y_off]? [font]? [font_size]? [bg_color]? [label_color]? [?]?;\r\n
    def __init__(self, x, y, **kwargs):
        PdCoreGui.__init__(self, "cnv", x, y, [], **kwargs)
        self._size = int(kwargs.get("size", 15))
        self.width = int(kwargs.get("width", 100))
        self.height = int(kwargs.get("height", 60))

    @staticmethod
    def from_atoms(atoms):
        assert len(atoms) == 13
        return PdGCanvas(0, 0,
                         size=atoms[0],
                         width=atoms[1],
                         height=atoms[2],
                         send=atoms[3],
                         receive=atoms[4],
                         label=atoms[5],
                         label_xoff=atoms[6],
                         label_yoff=atoms[7],
                         font_type=atoms[8],
                         font_size=atoms[9],
                         bg_color=atoms[10],
                         label_color=atoms[11])

    def draw(self, painter):
        painter.draw_rect(self.x, self.y, self.width, self.height, fill=self.bgcolor())
        self.draw_label(painter)


#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

 
__author__ = 'Serge Poltavski'

from pdcoregui import *


class PdRadio(PdCoreGui):
    def __init__(self, name, x, y, **kwargs):
        PdCoreGui.__init__(self, name, x, y, [], **kwargs)
        self._size = int(kwargs.get("size", 15))
        self._newold = int(kwargs.get("newold", 1))
        self._init = int(kwargs.get("init", 0))
        self._number = int(kwargs.get("number", 8))
        self._value = int(kwargs.get("value", 0))
        self._sq_offset = 3

    @staticmethod
    def from_atoms(atoms):
        assert len(atoms) == 16
        assert atoms[0] in ("vradio", "hradio")

        if atoms[0] == "vradio":
            return PdVRadio(0, 0,
                            size=atoms[1],
                            newold=atoms[2],
                            init=atoms[3],
                            number=atoms[4],
                            send=atoms[5],
                            receive=atoms[6],
                            label=atoms[7],
                            label_xoff=atoms[8],
                            label_yoff=atoms[9],
                            font_type=atoms[10],
                            font_size=atoms[11],
                            bg_color=atoms[12],
                            fg_color=atoms[13],
                            label_color=atoms[14],
                            default_value=atoms[15]
                            )
        else:
            return PdHRadio(0, 0,
                            size=atoms[1],
                            number=atoms[4])

class PdHRadio(PdRadio):
    def __init__(self, x, y, **kwargs):
        PdRadio.__init__(self, "hradio", x, y, **kwargs)

    def get_height(self):
        return self._size

    def get_width(self):
        return self._size * self._number

    def draw(self, painter):
        self.draw_bbox(painter)

        # draw cells
        assert self._number > 0
        step = self.width / self._number
        for i in range(1, self._number):
            x = self.x + i * step
            painter.draw_line(x, self.top, x, self.bottom, color=(0, 0, 0))

        # draw switch
        sx = self.x + self._size * self._value + self._sq_offset
        sy = self.y + self._sq_offset
        sz = self._size - 2 * self._sq_offset
        painter.draw_rect(sx, sy, sz, sz, fill=self.fgcolor())

        self.draw_label(painter)
        self.draw_xlets(painter)


class PdVRadio(PdRadio):
    def __init__(self, x, y, **kwargs):
        PdRadio.__init__(self, "vradio", x, y, **kwargs)

    def get_width(self):
        return self._size

    def get_height(self):
        return self._size * self._number

    def draw(self, painter):
        self.draw_bbox(painter)

        # draw cells
        assert self._number > 0
        step = self.height / self._number
        for i in range(1, self._number):
            y = self.y + i * step
            painter.draw_line(self.left, y, self.right, y, color=(0, 0, 0))

        # draw switch
        sy = self.y + self._size * self._value + self._sq_offset
        sx = self.x + self._sq_offset
        sz = self._size - 2 * self._sq_offset
        painter.draw_rect(sx, sy, sz, sz, fill=self.fgcolor())

        self.draw_label(painter)
        self.draw_xlets(painter)


#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2014 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

import pdobject
import re


class XletCalcDatabase(object):
    XLET_MESSAGE = pdobject.PdBaseObject.XLET_MESSAGE
    XLET_SOUND = pdobject.PdBaseObject.XLET_SOUND

    re_num = re.compile(r"^\d+$")

    def __init__(self):
        # INLETS
        # MESSAGE
        self._one_msg_inlet = (
            "bang", "b", "change", "makefilename", "print",
            "mtof", "ftom", "powtodb", "dbtopow", "rmstodb", "dbtorms",
            "sin", "cos", "tan", "atan", "atan2", "sqrt", "log", "exp", "abs",
            "loadbang", "bang~",
            "r~", "receive~", "unpack", "value", "v", "t", "trigger",
            "tabread", "tabread4", "soundfiler", "netsend", "qlist",
            "textfile", "openpanel", "savepanel", "init", "outlet"
        )

        self._two_msg_inlet = (
            "pipe", "i", "int", "float", "f", "symbol",
            "+", "-", "*", "/", "pow", "%",
            "==", "!=", "<", ">", ">=", "<=",
            "&", "&&", "|", "||", "mod", "div", "min", "max",
            "route", "spigot", "moses", "until", "swap", "delay", "del", "metro",
            "timer", "cputime", "realtime", "random", "pgmout", "bendout",
            "touchout", "midiout", "stripnote", "tabwrite", "bag", "poly",
        )

        self._tree_msg_inlet = (
            "clip", "line", "noteout", "ctlout", "polytouchout", "makenote"
        )
        # SOUND
        self._one_snd_inlet = (
            "s~", "send~", "sin~", "cos~", "q8_sqrt~", "q8_rsqrt~", "wrap~",
            "rfft~", "mtof~", "ftom~", "rmstodb~", "dbtorms~", "outlet~"
        )

        self._two_snd_inlet = (
            "rifft~", "fft~", "ifft~", "framp~"
        )

        # OUTLETS
        # MESSAGE
        self._one_msg_outlet = (
            "bang", "b", "float", "f", "int", "i", "symbol", "receive", "r",
            "change", "makefilename", "pipe", "+", "-", "*", "/", "%", "pow",
            "==", "!=", "<", ">", ">=", "<=", "&", "&&", "|", "||", "mtof",
            "ftom", "powtodb", "dbtopow", "rmstodb", "dbtorms",
            "sin", "cos", "tan", "atan", "atan2", "sqrt", "log", "exp", "abs",
            "random", "mod", "div", "min", "max", "clip", "loadbang", "bang~",
            "pack", "spigot", "until", "value", "v", "line", "delay", "del",
            "metro", "timer", "cputime", "realtime", "tabread", "tabread4",
            "soundfiler", "netsend", "openpanel", "savepanel", "bag", "key",
            "keyup", "init"
        )

        self._two_msg_outlet = (
            "moses", "swap", "sysexin", "midiin", "makenote", "stripnote",
            "netreceive", "qlist", "textfile", "keyname"
        )

        self._three_msg_outlet = {
            "poly",
        }

        # SOUND
        self._one_snd_outlet = (
            "sin~", "cos~", "osc~", "r~", "receieve~", "+~", "*~", "-~", "/~",
            "max~", "min~", "clip~", "q8_sqrt~", "q8_rsqrt~", "wrap~", "mtof~",
            "ftom~", "rmstodb~", "dbtorms~"
        )

        self._two_snd_outlet = (
            "fft~", "ifft~", "rfft~", "rifft~", "framp~"
        )

    def inlets(self, obj):
        if not issubclass(obj.__class__, pdobject.PdObject):
            return []

        name = obj.name
        nargs = obj.num_args()

        # 1 msg inlet
        if name in self._one_msg_inlet:
            return [self.XLET_MESSAGE]

        # 2 msg inlets
        if name in self._two_msg_inlet:
            return [self.XLET_MESSAGE] * 2

        # 3 msg inlets
        if name in self._tree_msg_inlet:
            return [self.XLET_MESSAGE] * 3

        # 1 snd inlet
        if name in self._one_snd_inlet:
            return [self.XLET_SOUND]

        # 2 snd inlet
        if name in self._two_snd_inlet:
            return [self.XLET_SOUND] * 2

        # [osc~]
        if name in ("osc~", ):
            return [self.XLET_SOUND, self.XLET_MESSAGE]

        return self.inlet_conditional(name, nargs)

    def inlet_conditional(self, name, nargs):
        # [s] or [send]
        if name in ("s", "send"):
            if nargs == 0:
                return [self.XLET_MESSAGE] * 2
            else:
                return [self.XLET_MESSAGE]

        # [dac~]
        if name == "dac~":
            if nargs == 0:
                return [self.XLET_SOUND] * 2
            else:
                return [self.XLET_SOUND] * nargs

        # 2 sound
        if name in ("*~", "-~", "+~", "/~", "max~", "min~"):
            if nargs == 0:
                return [self.XLET_SOUND] * 2
            else:
                return [self.XLET_SOUND, self.XLET_MESSAGE]

        if name in ("sel", "select", "route"):
            if nargs in (0, 1):
                return [self.XLET_MESSAGE] * 2
            else:
                return [self.XLET_MESSAGE]

        if name in ("pack",):
            if nargs == 0:
                return [self.XLET_MESSAGE] * 2
            else:
                return [self.XLET_MESSAGE] * nargs

        if name == "clip~":
            return [self.XLET_SOUND] + [self.XLET_MESSAGE] * 2

        if XletCalcDatabase.re_num.search(name):
            return [self.XLET_MESSAGE] * 2

        return []

    def outlets(self, obj):
        name = obj.name
        nargs = obj.num_args()

        # 1 msg outlet
        if name in self._one_msg_outlet:
            return [self.XLET_MESSAGE]

        # 2 msg outlet
        if name in self._two_msg_outlet:
            return [self.XLET_MESSAGE] * 2

        if name in self._three_msg_outlet:
            return [self.XLET_MESSAGE] * 3

        # 1 snd outlet
        if name in self._one_snd_outlet:
            return [self.XLET_SOUND]

        # 2 snd outlet
        if name in self._two_snd_outlet:
            return [self.XLET_SOUND] * 2

        return self.outlet_conditional(name, nargs)

    def outlet_conditional(self, name, nargs):
        lout_msg = lambda x, func: [self.XLET_MESSAGE] * func(x)

        if name in ("select", "sel", "route"):
            return lout_msg(nargs, lambda x: 2 if x == 0 else x + 1)

        if name in ("unpack", "t", "trigger"):
            return lout_msg(nargs, lambda x: 2 if x == 0 else x)

        if name in ("notein",):
            return lout_msg(nargs, lambda x: 3 if x == 0 else 2)

        if name in ("ctlin",):
            return lout_msg(nargs, lambda x: 3 if x == 0 else 2 if x == 1 else 1)

        if name in ("pgmin", "bendin", "touchin"):
            return lout_msg(nargs, lambda x: 2 if x == 0 else 1)

        if name in ("polytouchin",):
            return lout_msg(nargs, lambda x: 3 if x == 0 else 2)

        if XletCalcDatabase.re_num.search(name):
            return [self.XLET_MESSAGE]

        return []


#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #
 
__author__ = 'Serge Poltavski'

from pdcoregui import PdCoreGui

[size]? - number of digits the element displays
[height]? - vertical size of element in pixels
[min]? - minimum value, typically -1e+037
[max]? - maximum value, typically 1e+037
[log]? - linear when unset, logarithmic when set
[init]? - when set outputs
[send]? - send symbol name
[receive]? - receive symbol name
[label]? - label
[x_off]? - horizontal position of the label text relative to the upperleft corner of the object
[y_off]? - vertical position of the label text relative to the upperleft corner of the object
[font]? - font type
[fontsize]? - font size in pixels
[bg_color]? - background color
[fg_color]? - foreground color
[label_color]? - label color
[log_height]? - logarithmic steps, accepts values from 10 to 2000, default is 256


class PdNbx(PdCoreGui):
    def __init__(self, x, y, **kwargs):
        PdCoreGui.__init__(self, "nbx", x, y, [1])
        self._size = kwargs.get('size', )



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

from cairopainter import *
from pdobject import *
from brectcalculator import *
import argparse
import os.path


def detect_format(args):
    if args['format'] is None:
        name, ext = os.path.splitext(args['output'].lower())
        if len(ext) > 2:
            ext = ext[1:]
        else:
            ext = None
    else:
        ext = args['format'][0].lower()

    if ext not in ('png', 'pdf', 'svg'):
        raise ValueError("output format not supported or can't be detected")

    return ext

def draw_object(object, fname, format, **kwargs):
    w, h = BRectCalculator().object_brect(object)[2:]
    pad = 1 # pixel
    w += pad
    h += pad

    painter = CairoPainter(int(w), int(h), fname, format, **kwargs)
    painter.draw_object(object)

def main():
    arg_parser = argparse.ArgumentParser(description='PureData patch to image converter')
    arg_parser.add_argument('--format', '-f', metavar='format', nargs=1, choices=("png", "pdf", "svg"),
                            help='output image format')
    arg_parser.add_argument('name', metavar='OBJECT_NAME', help="PureData object name, for ex.: osc~")
    arg_parser.add_argument('output', metavar='OUTNAME', nargs='?', default='',
                            help="Image output name, for ex.: float.png")

    args = vars(arg_parser.parse_args())

    pdo = PdObject(args['name'])

    if not args['output']:
        if args['format']:
            output = "{0:s}.{1:s}".format(args['name'], args['format'][0].lower())
        else:
            output = args['name'] + ".png"
            args['output'] = output
    else:
        output = args['output']

    fmt = detect_format(args)
    draw_object(pdo, output, fmt)


if __name__ == '__main__':
    main()


#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2014 by Serge Poltavski                                   #
# serge.poltavski@gmail.com                                               #
# #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

from pdbaseobject import *
from pdpainter import *
import cairo
import textwrap
from pddrawstyle import *
from math import pi
from pdcanvas import PdCanvas


class CairoPainter(PdPainter):
    def __init__(self, width, height, output, fmt="png", **kwargs):
        PdPainter.__init__(self)
        self.style = PdDrawStyle()

        self.st_font_slant = cairo.FONT_SLANT_NORMAL
        self.st_font_weight = cairo.FONT_WEIGHT_NORMAL
        self.st_line_join = cairo.LINE_JOIN_ROUND

        assert width > 0
        assert height > 0

        self.width = width
        self.height = height

        assert fmt
        fmt = fmt.lower()
        self.format = fmt

        assert output
        self.output = output

        if fmt == 'png':
            self.ims = cairo.ImageSurface(cairo.FORMAT_ARGB32, self.width, self.height)
        elif fmt == 'pdf':
            assert self.output
            self.ims = cairo.PDFSurface(self.output, self.width, self.height)
        elif fmt == 'svg':
            assert self.output
            self.ims = cairo.SVGSurface(self.output, self.width, self.height)

        self.cr = cairo.Context(self.ims)

        self.set_src_color(self.style.fill_color)
        self.cr.rectangle(0, 0, self.width, self.height)
        self.cr.fill()

        self.cr.select_font_face(self.style.font_family, self.st_font_slant, self.st_font_weight)
        self.cr.set_font_size(self.style.font_size)
        self.cr.set_line_join(self.st_line_join)

        matrix = cairo.Matrix()
        if 'xoffset' in kwargs:
            matrix.translate(kwargs['xoffset'], 0)
        if 'yoffset' in kwargs:
            matrix.translate(0, kwargs['yoffset'])

        self.cr.set_matrix(matrix)

    def draw_rect(self, x, y, w, h, **kwargs):
        self.cr.save()
        if 'color' in kwargs:
            self.set_src_color(kwargs['color'])

        if 'width' in kwargs:
            self.cr.set_line_width(kwargs['width'])

        self.cr.rectangle(x + 0.5, y + 0.5, w, h)

        if 'fill' in kwargs:
            self.set_src_color(kwargs['fill'])
            self.cr.fill_preserve()

        self.cr.stroke()
        self.cr.restore()

    def set_src_color(self, rgb):
        if isinstance(rgb, float):
            self.cr.set_source_rgb(rgb, rgb, rgb)
        elif len(rgb) == 3:
            self.cr.set_source_rgb(rgb[0], rgb[1], rgb[2])

    def __del__(self):
        if self.format == 'png':
            self.ims.write_to_png(self.output)

    def draw_box(self, x, y, w, h):
        self.cr.save()
        self.cr.set_line_width(self.style.obj_line_width)
        self.set_src_color(self.style.obj_border_color)
        self.cr.rectangle(x + 0.5, y + 0.5, w, h)
        self.cr.stroke_preserve()
        self.set_src_color(self.style.obj_fill_color)
        self.cr.fill()
        self.cr.restore()

    def draw_txt(self, x, y, txt):
        (tx, ty, width, height, dx, dy) = self.cr.text_extents(txt)
        self.set_src_color(self.style.obj_text_color)
        self.cr.move_to(x + self.style.obj_pad_x, y + height + self.style.obj_pad_y)
        self.cr.show_text(txt)

    def draw_xlets(self, xlets, x, y, obj_width):
        if not xlets:
            return

        inlet_space = 0
        if len(xlets) > 1:
            inlet_space = (obj_width - len(xlets) * self.style.xlet_width) / (len(xlets) - 1)

        for num in xrange(0, len(xlets)):
            inx = x + num * inlet_space
            if num != 0:
                inx += num * self.style.xlet_width

            xlet = xlets[num]

            if xlet in (PdBaseObject.XLET_MESSAGE, PdBaseObject.XLET_GUI):
                self.set_src_color(self.style.xlet_msg_color)
            else:
                self.set_src_color(self.style.xlet_snd_color)

            iny = y
            self.cr.set_line_width(self.style.obj_line_width)
            self.cr.rectangle(inx + 0.5, iny + 0.5, self.style.xlet_width, self.xlet_height(xlet))
            self.cr.stroke_preserve()

            if xlet == PdBaseObject.XLET_SOUND:
                self.cr.fill()
                self.cr.stroke()
            else:
                self.cr.stroke()

            num += 1

    def draw_subpatch(self, subpatch):
        txt = "pd " + subpatch.name
        (x, y, width, height, dx, dy) = self.cr.text_extents(txt)
        x = subpatch.x
        y = subpatch.y
        w = width + self.style.obj_pad_x * 2
        h = self.style.obj_height

        self.draw_box(x, y, w, h)
        self.draw_txt(x, y, txt)

        self.draw_xlets(subpatch.inlets(), x, y, w)
        self.draw_xlets(subpatch.outlets(), x, y + h - self.style.xlet_msg_height, w)

    def draw_object(self, obj):
        txt = obj.to_string()
        (x, y, width, height, dx, dy) = self.cr.text_extents(txt)
        x = obj.x
        y = obj.y
        w = max(width + self.style.obj_pad_x * 2, self.style.obj_min_width)
        h = self.style.obj_height
        obj.set_width(w)
        obj.set_height(h)

        if hasattr(obj, "highlight"):
            pad = self.style.highlight_padding
            self.cr.rectangle(x + 0.5 - pad, y + 0.5 - pad, w + pad*2, h + pad*2)
            self.set_src_color(self.style.highlight_color)
            self.cr.fill()

        self.cr.set_line_width(self.style.obj_line_width)
        self.draw_box(x, y, w, h)
        self.draw_txt(x, y, txt)
        self.draw_xlets(obj.inlets(), x, y, w)
        self.draw_xlets(obj.outlets(), x, y + h - self.style.xlet_msg_height, w)

    def draw_message(self, message):
        txt = message.to_string()
        (x, y, width, height, dx, dy) = self.cr.text_extents(txt)
        x = message.x
        y = message.y
        w = width + self.style.obj_pad_x * 4
        h = self.style.obj_height
        message.set_height(h)
        message.set_width(w)

        # draw message box
        cr = self.cr
        self.set_src_color(self.style.obj_border_color)
        edge_w = 8
        edge_h = 10

        cr.save()
        cr.set_line_width(self.style.obj_line_width)
        tx = x + 0.5
        ty = y + 0.5
        cr.move_to(tx, ty)
        cr.line_to(tx, ty + h)
        cr.line_to(tx + w, ty + h)
        cr.line_to(tx + w - edge_w / 2.0, ty + (h + edge_h) / 2.0)
        cr.line_to(tx + w - edge_w / 2.0, ty + (h - edge_h) / 2.0)
        cr.line_to(tx + w, ty)
        cr.line_to(tx, ty)
        cr.stroke_preserve()
        self.set_src_color(self.style.msg_fill_color)
        cr.fill()
        cr.stroke()
        cr.restore()

        self.draw_txt(x, y, txt)
        self.draw_xlets(message.inlets(), x, y, w)
        self.draw_xlets(message.outlets(), x, y + h - self.style.xlet_msg_height, w)

    def xlet_height(self, t):
        if t == PdBaseObject.XLET_GUI:
            return self.style.xlet_gui_height
        else:
            return self.style.xlet_msg_height

    def inlet_connection_coords(self, obj, inlet_no):
        inlets = obj.inlets()

        xlet_space = 0
        if len(inlets) > 1:
            xlet_space = (obj.width - len(inlets) * self.style.xlet_width) / float(len(inlets) - 1)

        x = obj.x + (xlet_space + self.style.xlet_width) * inlet_no + self.style.xlet_width / 2.0
        y = obj.y

        return x, y

    def outlet_connection_coords(self, obj, outlet_no):
        outlets = obj.outlets()

        xlet_space = 0
        if len(outlets) > 1:
            xlet_space = (obj.width() - len(outlets) * self.style.xlet_width) / float(len(outlets) - 1)

        x = obj.x + (xlet_space + self.style.xlet_width) * outlet_no + self.style.xlet_width / 2.0
        if isinstance(obj, PdCanvas):
            y = obj.top + self.style.obj_height
        else:
            y = obj.bottom
        return x, y

    def draw_connections(self, canvas):
        for key, c in canvas.connections.items():
            src_obj = c[0]
            src_outl = c[1]
            if not src_obj.outlets():
                continue

            src_outl_type = src_obj.outlets()[src_outl]
            dest_obj = c[2]
            dest_inl = c[3]
            if not dest_obj.inlets():
                continue

            dest_inl_type = dest_obj.inlets()[dest_inl]

            sx, sy = self.outlet_connection_coords(src_obj, src_outl)
            dx, dy = self.inlet_connection_coords(dest_obj, dest_inl)

            # print sx, sy, "->", dx, dy

            self.cr.save()
            if dest_inl_type == PdBaseObject.XLET_SOUND and src_outl_type == PdBaseObject.XLET_SOUND:
                self.cr.set_line_width(self.style.conn_snd_width)
                self.set_src_color(self.style.conn_snd_color)
                # pixel correction
                if self.style.conn_snd_width % 2 == 0:
                    sx += 0.5
                    dx += 0.5

                self.cr.move_to(sx, sy)
                self.cr.line_to(dx, dy)
                self.cr.stroke()
                self.cr.set_dash(self.style.conn_snd_dash)
                self.set_src_color(self.style.conn_snd_color2)
            else:
                self.cr.set_line_width(self.style.conn_msg_width)
                self.set_src_color(self.style.conn_msg_color)
                # pixel correction
                if self.style.conn_msg_width % 2 == 1 and (sx - int(sx) == 0):
                    sx += 0.5
                    dx += 0.5

                sy += self.style.conn_msg_width

            self.cr.move_to(sx, sy)
            self.cr.line_to(dx, dy)
            self.cr.stroke()
            self.cr.restore()

        pass

    def draw_comment(self, comment):
        txt = comment.text()
        lines = textwrap.wrap(txt, 59)

        self.set_src_color(self.style.comment_color)
        lnum = 0
        for line in lines:
            line = ";\n".join(line.split(";")).split("\n")
            for subl in line:
                subl = subl.strip()
                self.draw_txt(comment.x, comment.y + lnum * self.style.font_size, subl)
                lnum += 1

    def draw_poly(self, vertexes, **kwargs):
        assert len(vertexes) > 1
        assert len(vertexes[0]) == 2
        self.cr.save()
        self.cr.set_line_width(self.style.obj_line_width)
        self.cr.move_to(vertexes[0][0] + 0.5, vertexes[0][1] + 0.5)
        for v in vertexes[1:]:
            self.cr.rel_line_to(v[0], v[1])

        self.cr.line_to(vertexes[0][0] + 0.5, vertexes[0][1] + 0.5)

        if 'fill' in kwargs:
            self.set_src_color(kwargs['fill'])
            self.cr.fill_preserve()

        if 'outline' in kwargs:
            self.set_src_color(kwargs['outline'])
            self.cr.stroke()

        self.cr.restore()

    def draw_text(self, x, y, text, **kwargs):
        self.cr.save()

        if 'color' in kwargs:
            self.set_src_color(kwargs['color'])
        else:
            self.set_src_color((0, 0, 0))

        if 'font_size' in kwargs:
            self.cr.set_font_size(kwargs['font_size'])

        if 'font' in kwargs:
            self.cr.select_font_face(kwargs['font'])

        self.cr.move_to(x, y)
        self.cr.show_text(text)
        self.cr.restore()

    def draw_inlets(self, inlets, x, y, width):
        self.draw_xlets(inlets, x, y, width)

    def draw_outlets(self, outlets, x, y, width):
        if not outlets:
            return

        yoffset = self.style.xlet_height(outlets[0])
        self.draw_xlets(outlets, x, y - yoffset, width)

    def draw_circle(self, x, y, radius, **kwargs):
        self.cr.save()
        self.cr.set_line_width(self.style.obj_line_width)
        self.cr.arc(x, y, radius, 0, 2*pi)
        if 'fill' in kwargs:
            self.set_src_color(kwargs['fill'])
            self.cr.fill_preserve()

        if 'outline' in kwargs:
            self.set_src_color(kwargs['outline'])
            self.cr.stroke()

        self.cr.restore()

    def draw_line(self, x0, y0, x1, y1, **kwargs):
        self.cr.save()
        self.cr.set_line_width(float(kwargs.get('width', 1)))
        if 'color' in kwargs:
            self.set_src_color(kwargs['color'])

        self.cr.set_line_cap(cairo.LINE_CAP_BUTT)
        self.cr.move_to(x0 + 0.5, y0 + 0.5)
        self.cr.line_to(x1 + 0.5, y1 + 0.5)
        self.cr.stroke()
        self.cr.restore()

    def draw_graph(self, graph):
        print graph


#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

 
__author__ = 'Serge Poltavski'

from pdcoregui import PdCoreGui

class PdToggle(PdCoreGui):
    @staticmethod
    def from_atoms(atoms):
        return PdToggle(0, 0,
                        size=atoms[0],
                        init=atoms[1],
                        send=atoms[2],
                        receive=atoms[3],
                        label=atoms[4],
                        label_xoff=atoms[5],
                        label_yoff=atoms[6],
                        font_type=atoms[7],
                        font_size=atoms[8],
                        bg_color=atoms[9],
                        fg_color=atoms[10],
                        label_color=atoms[11],
                        init_value=atoms[12],
                        default_value=atoms[13])

    def __init__(self, x, y, **kwargs):
        # tgl_defaults = ['15', '0', 'empty', 'empty', 'empty', '17', '7', '0', '10', '-262144', '-1', '-1', '0', '1']
        PdCoreGui.__init__(self, "tgl", x, y, [], **kwargs)
        # square size of the gui element
        self._size = int(kwargs.get("size", 15))
        self._init = int(kwargs.get("init", 0))
        # value sent when the [init] attribute is set
        self._init_value = int(kwargs.get("init_value", 0))
        # default_value when the [init]? attribute is not set
        self._default_value = int(kwargs.get("default_value", 1))
        self._checked = kwargs.get("checked", False)

    def get_width(self):
        return self._size

    def get_height(self):
        return self._size

    def draw(self, painter):
        self.draw_bbox(painter)

        if self._checked:
            painter.draw_line(self.left, self.top, self.right, self.bottom, color=self.fgcolor())
            painter.draw_line(self.left, self.bottom, self.right, self.top, color=self.fgcolor())

        self.draw_label(painter)
        self.draw_xlets(painter)


#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #


__author__ = 'Serge Poltavski'

from pdcoregui import *
import six


class PdBng(PdCoreGui):
    # default arguments:
    # X obj 256 25 bng 15 250 50 0 empty empty empty 17 7 0 10 -262144 -1 -1;

    def __init__(self, x, y, **kwargs):
        PdCoreGui.__init__(self, "bng", x, y, [], **kwargs)
        self._size = int(kwargs.get("size", 15))
        assert self._size > 0

        self._hold = int(kwargs.get("hold", 250))
        assert 50 <= self._hold <= 1000000000

        self._interrupt = int(kwargs.get("interrupt", 50))
        assert 10 <= self._interrupt <= 250

        self._init = int(kwargs.get("init", 0))
        self._pressed = 'pressed' in kwargs and kwargs['pressed'].lower() == "true"

    @staticmethod
    def from_atoms(atoms):
        assert isinstance(atoms, list)
        assert len(atoms) == 14

        return PdBng(0, 0,
                     size=int(atoms[0]),
                     hold=int(atoms[1]),
                     interrupt=int(atoms[2]),
                     init=int(atoms[3]),
                     send=atoms[4],
                     receive=atoms[5],
                     label=atoms[6],
                     label_xoff=atoms[7],
                     label_yoff=atoms[8],
                     font_type=atoms[9],
                     font_size=atoms[10],
                     bg_color=atoms[11],
                     fg_color=atoms[12],
                     label_color=atoms[13])

    def get_height(self):
        return self._size

    def get_width(self):
        return self._size

    def center(self):
        return self.x + self.width/2 + 0.75, self.y + self.height/2 + 0.75

    def draw(self, painter):
        self.draw_bbox(painter)
        cx, cy = self.center()

        circle_color = self.bgcolor()
        if self._pressed:
            circle_color = self.fgcolor()
        painter.draw_circle(cx, cy, (self.width-1.5)/2, fill=circle_color)

        self.draw_label(painter)
        self.draw_xlets(painter)




#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2014 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'
import re


class PdBaseObject(object):
    brect_calculator = None
    XLET_MESSAGE, XLET_SOUND, XLET_GUI = range(0, 3)

    def __init__(self, x=0, y=0, w=0, h=0):
        self._id = -1
        self._x = int(x)
        self._y = int(y)
        self._height = int(h)
        self._width = int(w)

    x = property(lambda self: self.get_x(), lambda self, v: self.set_x(v))
    y = property(lambda self: self.get_y(), lambda self, v: self.set_y(v))
    height = property(lambda self: self.get_height(), lambda self, v: self.set_height(v))
    width = property(lambda self: self.get_width(), lambda self, v: self.set_width(v))

    def is_null(self):
        return not self.x and not self.y and not self.width and not self.height

    def brect(self):
        return self.x, self.y, self.width, self.height

    def move_by(self, x, y):
        self.x += x
        self.y += y

    @property
    def top(self):
        return self.y

    @property
    def bottom(self):
        return self.y + self.height

    @property
    def left(self):
        return self.x

    @property
    def right(self):
        return self.x + self.width

    def get_x(self):
        return self._x

    def set_x(self, x):
        self._x = int(x)

    def get_y(self):
        return self._y

    def set_y(self, y):
        self._y = int(y)

    def get_height(self):
        return self._height

    def set_height(self, h):
        self._height = int(h)

    def get_width(self):
        return self._width

    def set_width(self, w):
        self._width = int(w)

    def draw(self, painter):
        assert not "Not implemented"

    def inlets(self):
        return ()

    def outlets(self):
        return ()

    def traverse(self, visitor):
        assert not "Not implemented"

    def calc_brect(self):
        self.brect_calculator.calc(self)

    @staticmethod
    def unescape(string):
        return re.sub(r'\\(.)', r'\1', string)

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, oid):
        self._id = int(oid)


#!/usr/bin/env python
# coding=utf-8

from setuptools import setup


def readme():
    with open('README.md') as f:
        return f.read()


setup(name='pddoc',
      version='0.1.4',
      description='PureData documentation tools',
      long_description=readme(),
      classifiers=[
          'Development Status :: 2 - Pre-Alpha',
          'Topic :: Software Development :: Documentation',
          'Topic :: Multimedia :: Sound/Audio',
          'Programming Language :: Python :: 2.7',
          'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)'
      ],
      url='https://github.com/uliss/pddoc',
      author='Serge Poltavski',
      author_email='serge.poltavski@gmail.com',
      keywords='puredata documentation',
      license='GPLv3',
      packages=['pddoc'],
      install_requires=[
          'termcolor', 'colorama', 'six', 'mako', 'argparse'
      ],
      test_suite='nose.collector',
      tests_require=['nose'],
      entry_points={
          'console_scripts': ['pd2image=pddoc.pd2image:main']
      },
      include_package_data=True,
      zip_safe=False)



#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2014 by Serge Poltavski                                   #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #


__author__ = 'Serge Poltavski'

from os import path
import re
import common
from pdcanvas import *
from pdmessage import *
from pdcomment import *
from pdcoregui import *
from pdarray import *
import pdfactory


class PdParser:
    lines_re = re.compile("(#(.*?)[^\\\])\r?\n?;\r?\n", re.MULTILINE | re.DOTALL)
    split_re = re.compile(" |\r\n?|\n", re.MULTILINE)

    def __init__(self):
        self.canvas = None
        self.canvas_stack = []
        self._array = None

    def parse_canvas(self, atoms):
        # get positions
        values = []
        for i in xrange(3, -1, -1):
            if len(atoms) > i:
                values.insert(0, int(atoms[i]))
            else:
                values.insert(0, 0)

        x, y, w, h = values

        # get canvas name
        kwargs = {}
        if len(atoms) >= 4:
            kwargs['name'] = atoms[4]

        c = PdCanvas(x, y, w, h, **kwargs)

        # root canvas
        if self.canvas is None:
            c.type = PdCanvas.TYPE_WINDOW
            self.canvas = c
        else:
            c.type = PdCanvas.TYPE_SUBPATCH

        self.canvas_stack.append(c)

    def current_canvas(self):
        assert len(self.canvas_stack) > 0
        return self.canvas_stack[-1]

    def parse_frameset(self, atoms):
        if atoms[0] == "canvas":
            atoms.pop(0)
            self.parse_canvas(atoms)
        else:
            common.warning(u"unknown frameset type: {0:s}".format(atoms[0]))

    def parse_messages(self, atoms):
        x = atoms[0]
        y = atoms[1]
        atoms.pop(0)
        atoms.pop(0)

        msg = PdMessage(x, y, atoms)
        self.current_canvas().append_object(msg)

    def parse_comments(self, atoms):
        x = atoms[0]
        y = atoms[1]
        comment = PdComment(x, y, atoms[2:])
        self.current_canvas().append_object(comment)

    def parse_restore(self, atoms):
        cnv_type = atoms[2]
        if cnv_type == "graph":
            self.current_canvas().type = PdCanvas.TYPE_GRAPH
            c = self.canvas_stack.pop()
            self._array.x = int(atoms[0])
            self._array.y = int(atoms[1])
            self.canvas.append_object(self._array)
            self._array = None
        elif cnv_type == "pd":
            c = self.canvas_stack.pop()
            c.type = PdCanvas.TYPE_SUBPATCH
            c.x = int(atoms[0])
            c.y = int(atoms[1])
            self.current_canvas().append_subpatch(c)
        else:
            common.warning(u"unknown canvas type: {0:s}".format(cnv_type))
            assert False

    def parse_obj(self, atoms):
        x = atoms[0]
        y = atoms[1]

        obj = pdfactory.make(atoms[2:])
        obj.x = x
        obj.y = y

        self.current_canvas().append_object(obj)

    def parse_array(self, atoms):
        self._array = PdArray.from_atoms(atoms[1:])

    def parse_array_content(self, atoms):
        assert self._array
        self._array.set(atoms)

    def parse_connect(self, atoms):
        c = self.current_canvas()
        c.connect(atoms)

    def parse_coords(self, atoms):
        if not self._array:
            return

        self._array.set_xrange(float(atoms[1]), float(atoms[3]))
        self._array.set_yrange(float(atoms[2]), float(atoms[4]))
        self._array.width = int(atoms[5])
        self._array.height = int(atoms[6])

    def parse_objects(self, atoms):
        name = atoms[0]
        if name == 'msg':
            atoms.pop(0)
            self.parse_messages(atoms)
        elif name == "obj":
            atoms.pop(0)
            self.parse_obj(atoms)
        elif name == "connect":
            atoms.pop(0)
            self.parse_connect(atoms)
        elif name == "text":
            atoms.pop(0)
            self.parse_comments(atoms)
        elif name == "restore":
            # end canvas definition
            atoms.pop(0)
            self.parse_restore(atoms)
        elif name == "floatatom":
            obj = pdfactory.make(atoms)
            self.current_canvas().append_object(obj)
        elif name == "array":
            self.parse_array(atoms)
        elif name == "coords":
            self.parse_coords(atoms)
        else:
            print name

    def parse_atoms(self, atoms):
        if atoms[0] == '#X':
            atoms.pop(0)
            self.parse_objects(atoms)
        elif atoms[0] == '#N':
            atoms.pop(0)
            self.parse_frameset(atoms)
        elif atoms[0] == '#A':
            self.parse_array_content(atoms[2:])
        else:
            common.info("unknown token: {0:s}".format(atoms[0]))

    def parse(self, file_name):
        if not path.exists(file_name):
            common.warning(u"File not exists: \"{0:s}\"".format(file_name))
            return False

        f = open(file_name, "r")
        lines = f.read()
        f.close()

        for found in self.lines_re.finditer(lines):
            line = found.group(1)
            atoms = self.split_re.split(line)
            self.parse_atoms(atoms)

        return True


#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2014 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

from pdcanvas import *
import textwrap


class PdExporter(object):
    def __init__(self):
        self.result = []

    def visit_canvas_begin(self, cnv):
        assert isinstance(cnv, PdCanvas)

        if cnv.type == PdCanvas.TYPE_WINDOW:
            line = "#N canvas {0:d} {1:d} {2:d} {3:d} {4:s};". \
                format(cnv.x, cnv.y, cnv.width, cnv.height, cnv.name)
            self.result.append(line)
        elif cnv.type == PdCanvas.TYPE_SUBPATCH:
            line = "#N canvas {0:d} {1:d} {2:d} {3:d} {4:s} 0;". \
                format(cnv.x, cnv.y, cnv.width, cnv.height, cnv.name)
            self.result.append(line)
            pass

    def visit_graph(self, graph):
        pass

    def visit_subpatch(self, spatch):
        pass

    def visit_canvas_end(self, cnv):
        if cnv.type == PdCanvas.TYPE_WINDOW:
            self.result.append("")
        elif cnv.type == PdCanvas.TYPE_SUBPATCH:
            line = "#X restore {0:d} {1:d} pd {2:s};".format(cnv.x, cnv.y, cnv.name)
            self.result.append(line)
        else:
            pass

    def visit_connection(self, conn):
        line = "#X connect {0:d} {1:d} {2:d} {3:d};".format(conn[0].id, conn[1], conn[2].id, conn[3])
        self.result.append(line)

    def visit_comment(self, comment):
        line = "#X text {0:d} {1:d} {2:s};".format(comment.x, comment.y, " ".join(comment.args))
        for l in textwrap.wrap(line):
            self.result.append(l)

    def visit_object(self, obj):
        line = "#X obj {0:d} {1:d} {2:s}".format(obj.x, obj.y, obj.name)

        if len(obj.args):
            line += " " + " ".join(obj.args)

        line += ";"

        for l in textwrap.wrap(line, 70):
            self.result.append(l)

    def visit_core_gui(self, gui):
        self.visit_object(gui)

    def visit_message(self, msg):
        line = "#X msg {0:d} {1:d} {2:s};".format(msg.x, msg.y, msg.args_to_string())
        self.result.append(line)

    def save(self, fname):
        f = open(fname, 'w')
        f.write("\n".join(self.result))
        f.close()


#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2014 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

import common
import xml.etree.ElementTree as ET


def make_class_name(tag_name):
    return "Doc%s" % (tag_name.capitalize())


def create_instance(tag_name, *args):
    class_name = make_class_name(tag_name)
    obj = globals()[class_name](args)
    return obj


class DocItem(object):
    def __init__(self, *args):
        self._elements = []
        self._text = ""

        if len(args) > 1:
            self._text = args[0]

    def items(self):
        return self._elements

    def text(self):
        return self._text

    def is_valid_tag(self, tag_name):
        return False

    def add_child(self, element):
        self._elements.append(element)

    def clear(self):
        self._elements = []

    def num_children(self):
        return len(self._elements)

    def traverse(self, visitor):
        mname = self.__class__.__name__[3:].lower()
        method = mname + "_begin"
        if hasattr(visitor, method):
            getattr(visitor, method)(self)

        self.traverse_children(visitor)

        method = mname + "_end"
        if hasattr(visitor, method):
            getattr(visitor, method)(self)

    def traverse_children(self, visitor):
        for e in self._elements:
            e.traverse(visitor)

    def read_xml_data(self, xmlobj):
        self._text = xmlobj.text

    def from_xml(self, xmlobj):
        self.read_xml_data(xmlobj)

        for child in xmlobj:
            if not self.is_valid_tag(child.tag):
                common.warning("tag <%s> not allowed in <%s>" % (child.tag, xmlobj.tag))
                continue
                # break

            obj = create_instance(child.tag)
            obj.from_xml(child)
            self.add_child(obj)


class DocTitle(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocMeta(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name in ("description", "authors", "contacts",
                            "license", "version", "website",
                            "aliases", "keywords", "library",
                            "also", "category")


class DocDescription(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocAuthors(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name in ("author",)


class DocAuthor(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocAlso(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name == "see"


class DocSee(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocContacts(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocLibrary(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocVersion(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocCategory(DocItem):
    pass


class DocLicense(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._url = ""
        self._name = ""

    def read_xml_data(self, xmlobj):
        self._url = xmlobj.attrib.get("url", "")
        self._name = xmlobj.text

    def url(self):
        return self._url

    def name(self):
        return self._name


class DocExample(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name in ("pdexample",)


class DocPdobject(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._id = ""
        self._name = ""
        self._args = ""
        self._comment = ""
        self._offset = 0
        self._highlight = False
        self._attrs = {}

    @property
    def id(self):
        return self._id

    def is_valid_tag(self, tag_name):
        return tag_name in ("pdinlet", "pdoutlet")

    def highlight(self):
        return self._highlight == "true"

    def name(self):
        return self._name

    def offset(self):
        return self._offset

    def from_xml(self, xmlobj):
        try:
            self._id = xmlobj.attrib["id"]
            self._name = xmlobj.attrib["name"]
            self._args = xmlobj.attrib.get("args", "")
            self._comment = xmlobj.attrib.get("comment", "")
            self._offset = int(xmlobj.attrib.get("offset", "0"))
            self._highlight = xmlobj.attrib.get("highlight", "")

            # save other attrs
            for k, v in xmlobj.attrib.items():
                if k not in ('id', 'name', 'args', 'comment', 'offset', 'highlight'):
                    self._attrs[k] = v

            DocItem.from_xml(self, xmlobj)
        except KeyError, e:
            common.warning("required attribute not found: \"%s\" in <%s>" % (e.message, xmlobj.tag))

    @property
    def comment(self):
        return self._comment

    def args(self):
        return self._args.split(" ")

    def attrs(self):
        return self._attrs


class DocPdmessage(DocPdobject):
    def __init__(self, *args):
        DocPdobject.__init__(self, args)

    def is_valid_tag(self, name):
        return False

    def from_xml(self, xmlobj):
        try:
            self._id = xmlobj.attrib["id"]
            self._comment = xmlobj.attrib.get("comment", "")
            self._offset = int(xmlobj.attrib.get("offset", "0"))
            DocItem.from_xml(self, xmlobj)
            # after parent method
            self._text = xmlobj.attrib["text"]
        except KeyError, e:
            common.warning("required attribute not found: \"%s\" in <%s>" % (e.message, xmlobj.tag))


class DocPdinlet(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocPdoutlet(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocRow(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name in ("col", "pdmessage", "pdobject", "pdcomment")


class DocCol(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name in ("row", "pdmessage", "pdobject", "pdcomment")


class DocPdconnect(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._src_id = ""
        self._dest_id = ""
        self._src_out = None
        self._dest_in = None

    def src_id(self):
        return self._src_id

    def dest_id(self):
        return self._dest_id

    def from_xml(self, xmlobj):
        self._src_out = int(xmlobj.attrib["src_out"])
        self._dest_in = int(xmlobj.attrib["dest_in"])
        self._dest_id = xmlobj.attrib["dest"]
        self._src_id = xmlobj.attrib["src"]
        DocItem.from_xml(self, xmlobj)


class DocPdcomment(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocPdexample(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._width = 0
        self._height = 0
        self._file = ""
        self._size = ""
        self._title = ""

    def is_valid_tag(self, tag_name):
        return tag_name in ("row", "col", "pdconnect")

    def from_xml(self, xmlobj):
        size = xmlobj.attrib.get("size", None)
        if size:
            if size in ("auto", "canvas"):
                self._size = size
            else:
                common.warning("invalid size value: \"{0:s}\" in <{1:s}>".format(size, xmlobj.tag))

        self._width = int(xmlobj.attrib.get("width", 0))
        self._height = int(xmlobj.attrib.get("height", 0))
        self._file = xmlobj.attrib.get("file", "")
        self._title = xmlobj.attrib.get("title", "")
        DocItem.from_xml(self, xmlobj)

    def file(self):
        return self._file

    def width(self):
        return self._width

    def height(self):
        return self._height

    def size(self):
        return self._size

    def title(self):
        return self._title


class DocWebsite(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)


class DocKeywords(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._keywords = []

    def from_xml(self, xmlobj):
        self._keywords = xmlobj.text.split(" ")
        super(self.__class__, self).from_xml(xmlobj)

    def keywords(self):
        return self._keywords


class DocAliases(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._aliases = []

    def aliases(self):
        return self._aliases

    def from_xml(self, xmlobj):
        for child in xmlobj:
            if child.tag == "alias":
                self._aliases.append(child.text)


class DocXlets(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)

    def xlet_dict(self):
        res = {}
        for inl in self.items():
            n = inl.number()
            if n not in res:
                res[n] = []

            res[n].append(inl)

        return res


class DocInlets(DocXlets):
    def __init__(self, *args):
        DocXlets.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name == "inlet"

    def inlet_dict(self):
        return self.xlet_dict()


class DoxTypeElement(DocItem):
    allowed_types = ("bang", "float", "list", "symbol", "pointer", "any")

    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._type = ""

    def is_valid_type(self, t):
        return t in self.allowed_types

    def type(self):
        return self._type

    def from_xml(self, xmlobj):
        self._type = xmlobj.attrib["type"]
        DocItem.from_xml(self, xmlobj)


class DocXlet(DoxTypeElement):
    def __init__(self, *args):
        DoxTypeElement.__init__(self, *args)
        self._maxvalue = ""
        self._minvalue = ""
        self._number = ""

    def number(self):
        return int(self._number)

    def from_xml(self, xmlobj):
        self._maxvalue = xmlobj.attrib.get("maxvalue", "")
        self._minvalue = xmlobj.attrib.get("minvalue", "")
        self._number = xmlobj.attrib["number"]
        DoxTypeElement.from_xml(self, xmlobj)

    def range(self):
        if not self._minvalue and not self._maxvalue:
            return ()

        return float(self._minvalue), float(self._maxvalue)


class DocInlet(DocXlet):
    def __init__(self, *args):
        DocXlet.__init__(self, args)


class DocOutlets(DocXlets):
    def __init__(self, *args):
        DocXlets.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name == "outlet"

    def outlet_dict(self):
        return self.xlet_dict()


class DocOutlet(DocXlet):
    def __init__(self, *args):
        DocXlet.__init__(self, args)


class DocArgument(DoxTypeElement):
    def __init__(self, *args):
        DoxTypeElement.__init__(self, args)


class DocArguments(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name == "argument"

    def argument_count(self):
        return len(self._elements)


class DocInfo(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name in ("itemize", "p")


class DocItemize(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)

    def is_valid_tag(self, tag_name):
        return tag_name in ("item",)


class DocObject(DocItem):
    def __init__(self, *args):
        DocItem.__init__(self, args)
        self._name = ""

    def is_valid_tag(self, tag_name):
        return tag_name in ("title", "meta", "inlets", "outlets", "arguments", "info", "example")

    def from_xml(self, xobj):
        self._name = xobj.attrib["name"]
        DocItem.from_xml(self, xobj)